#include <math.h>
#include <stdlib.h>
#include <stdio.h>
#include "path.h"
#include "option.h"
#include "mesh.h"
#include "blackScholes.h"
#include "util.h"

/* The exact value of control 1 used for the american estimate at time, path */
double innerControl1Exact(meshData *md, int time, int path){
  int max;

  max = dWhichMax(md->option.pData.underliers, md->option.pData.price[time][path]);

  return(exp(-time*md->option.riskFreeRate)*euroCall(md->option.pData.price[time][path][max],
						     md->option.strikePrice,
						     md->option.riskFreeRate,
						     sqrt(md->option.pData.covariance[max][max]),
						     1.0,
						     md->option.pData.dividendRate[max]));
}



/* The mesh estimate of control1 at time, path */
double innerControl1Estimate(meshData *md, int time, int path){
  int max, toPath;
  double estimate,deflatedPrice,weight,sumWeight;
  double e,p,r;
  
  max = dWhichMax(md->option.pData.underliers, md->option.pData.price[time][path]); 

  sumWeight = 0;
  estimate = 0;
  for(toPath=0; toPath<md->meshPaths; toPath++){
    sumWeight += md->weights[time][path][toPath];
    if(md->option.pData.price[time+1][toPath][max] > md->option.strikePrice){
      deflatedPrice = exp(-md->option.riskFreeRate)*(md->option.pData.price[time+1][toPath][max] - md->option.strikePrice);
    }else{
      deflatedPrice = 0;
    }
    estimate += md->weights[time][path][toPath]*deflatedPrice;
    /*    printf("E:%i\t%i\t%lf\t%lf\n",time,path,md->weights[time][path][toPath],deflatedPrice); */
  }
  /*estimate /= sumWeight;*/
  /*  printf("E:%lf\t%lf\n",sumWeight,exp(-time*md->option.riskFreeRate)*estimate);*/
  return(exp(-time*md->option.riskFreeRate)*estimate);
}



/* The variance of the mesh estimate of control1 at time,path */
double innerControl1Variance(meshData *md, int time, int path){
  int max, toPath;
  double estimate,sumWeight,deflatedPrice;

  max = dWhichMax(md->option.pData.underliers, md->option.pData.price[time][path]); 

  estimate = 0;
  sumWeight =0;
  for(toPath=0; toPath<md->meshPaths; toPath++){
    if(md->option.pData.price[time+1][toPath][max] > md->option.strikePrice){
      deflatedPrice = exp(-md->option.riskFreeRate) * (md->option.pData.price[time+1][toPath][max]-md->option.strikePrice);
    }else{
      deflatedPrice = 0;
    }
    estimate += md->weights[time][path][toPath] *deflatedPrice*deflatedPrice;
    sumWeight += md->weights[time][path][toPath];
    /*    printf("V:%i\t%i\t%lf\t%lf\n",time,path,md->weights[time][path][toPath],deflatedPrice);*/
  }
  /*estimate /= sumWeight;*/
  /*  printf("V:%lf\t%lf\n",sumWeight,estimate);*/
  return(exp(-2*time*md->option.riskFreeRate)*estimate);
}



/* Mesh estimate of american price.  Should probably be in mesh.c 
 * Assumes weights to be calculated.  This could be removed from 
 * meshDataFill and made a function in mesh.c
 */
double innerMeshEstimate(meshData *md, int time, int path){
int max, toPath;
  double estimate,sumWeight;
  double w,v;

  estimate = 0;
  sumWeight = 0;
  for(toPath=0; toPath<md->meshPaths; toPath++){
    sumWeight += md->weights[time][path][toPath];
    estimate +=  md->weights[time][path][toPath]*md->newValue[time+1][toPath];
    /*    printf("M:%i\t%i\t%i\t%lf\t%lf\n",time,path,toPath,md->weights[time][path][toPath],md->newValue[time+1][toPath]);*/
  }

  /*estimate /= sumWeight;*/
  /*  printf("M:%lf\t%lf\n",sumWeight,estimate);*/
  return(estimate);
} 



/* The correlation of the american estimate and control1 */
double innerControl1Correlation(meshData *md, int time, int path){
int max, toPath;
  double estimate,sumWeight,deflatedPrice;

  max = dWhichMax(md->option.pData.underliers, md->option.pData.price[time][path]); 

  estimate = 0;
  sumWeight = 0;
  for(toPath=0; toPath<md->meshPaths; toPath++){
    if(md->option.pData.price[time+1][toPath][max] > md->option.strikePrice){
      deflatedPrice = exp(-md->option.riskFreeRate) * (md->option.pData.price[time+1][toPath][max]-md->option.strikePrice);
    }else{
      deflatedPrice = 0;
    }
    estimate += md->weights[time][path][toPath] * exp(-time*md->option.riskFreeRate)*deflatedPrice*md->newValue[time+1][toPath];
    sumWeight += md->weights[time][path][toPath];
    /*    printf("C:%i\t%i\t%lf\t%lf\n",time,path,md->weights[time][path][toPath],exp(-time*md->option.riskFreeRate)*deflatedPrice*md->newValue[time+1][toPath]);*/
  }
  /*estimate /= sumWeight;*/
  /*  printf("C:%lf\t%lf\n",sumWeight,estimate);*/
  return(estimate);
} 


/* The controlled value of the american estimate at time, path */
double innerNewEstimate(meshData *md, int time, int path){
  double beta, I1, J1, J1V, J, I1J1, S;
  int toPath;

  S = 0;
  for(toPath=0;toPath<md->meshPaths;toPath++){
    S += md->weights[time][path][toPath];
  }

  I1   = innerMeshEstimate(md,time,path);
  J1   = innerControl1Estimate(md,time,path);
  J1V  = innerControl1Variance(md,time,path);
  J    = innerControl1Exact(md,time,path);
  I1J1 = innerControl1Correlation(md,time,path);

  beta = (S*I1J1 - I1 * J1)/(S*J1V - J1 * J1);

/*   printf("N:%i\t%i\t",time,path); */
/*   printf("%lf\t%lf\t%lf\t%lf\t%lf\t%lf\t%lf\n",I1,J1,J,J1V,I1J1,beta,I1 - beta*(J1 - J)); */
  return(I1 - beta*(J1 - J));
}

void innerDataFill(meshData *md){
  int time,path;

  for(time=md->option.pData.times-2;time>0;time--){
    for(path=md->option.pData.paths-1;path>=0;path--){
      md->newValueIfHeld[time][path] = innerNewEstimate(md,time,path);
      if(md->newValueIfHeld[time][path] > md->option.payoff[time][path]){
	md->newValue[time][path] = md->newValueIfHeld[time][path];
      }else{
	md->newValue[time][path] = md->option.payoff[time][path];
      }
    }
  }

  md->newValueIfHeld[0][0] = innerNewEstimate(md,0,0);
  if(md->newValueIfHeld[0][0] > md->option.payoff[0][0]){
    md->newValue[time][path] = md->newValueIfHeld[0][0];
  }else{
    md->newValue[0][0] = md->option.payoff[0][0];
  }
}

double innerControlledPathEstimate(meshData *md){
  int time, path;
  double value;

  if(md->option.payoff[0][0] >= md->newValueIfHeld[0][0]){
    return(md->option.payoff[0][0]);
  }

  value = 0;
  for(path=md->meshPaths; path < md->option.pData.paths; path++){
    for(time=1; time < md->option.pData.times; time++){
      if(md->option.payoff[time][path] > md->newValueIfHeld[time][path]){
	value += md->option.payoff[time][path];
	break;
      }
    }
  }
  return(value/((double)(md->option.pData.paths-md->meshPaths)));
}
